import logging
from pathlib import Path

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

logger = logging.getLogger(__name__)

configuration = {}


def get_configpath():
    for pn in (
        Path.home() / '.config' / 'vm-compact',
        Path.home() / 'Documents' / 'Voicemeeter' / 'vm-compact',
        Path.cwd() / '_internal' / 'configs',
        Path.cwd() / 'configs',
    ):
        if pn.exists():
            return pn


if configpath := get_configpath():
    filepaths = list(configpath.glob('*.toml'))
    if any(f.stem in ('app', 'vban') for f in filepaths):
        configs = {}
        for filepath in filepaths:
            filename = filepath.with_suffix('').stem
            if filename in ('app', 'vban'):
                try:
                    with open(filepath, 'rb') as f:
                        configs[filename] = tomllib.load(f)
                    logger.info(f'configuration: {filename} loaded into memory')
                except tomllib.TOMLDecodeError:
                    logger.error(f'Invalid TOML config: configs/{filename.stem}')
        configuration |= configs

_defaults = {
    'configs': {
        'config': None,
    },
    'theme': {
        'enabled': True,
        'mode': 'light',
    },
    'extends': {
        'extended': True,
        'extends_horizontal': True,
    },
    'channel': {
        'width': 80,
        'height': 130,
        'xpadding': 3,
    },
    'mwscroll_step': {
        'size': 3,
    },
    'submixes': {
        'default': 0,
    },
    'navigation': {'show': False},
}


if 'app' in configuration:
    for key in _defaults:
        if key in configuration['app']:
            configuration['app'][key] = _defaults[key] | configuration['app'][key]
        else:
            configuration['app'][key] = _defaults[key]
else:
    configuration['app'] = _defaults


def get_configuration(key):
    if key in configuration:
        return configuration[key]


def loader(kind_id, target):
    configs = {'reset': target.configs['reset']}
    if configpath := get_configpath():
        userconfigpath = configpath / kind_id
        if userconfigpath.exists():
            filepaths = list(userconfigpath.glob('*.toml'))
            for filepath in filepaths:
                identifier = filepath.with_suffix('').stem
                try:
                    with open(filepath, 'rb') as f:
                        configs[identifier] = tomllib.load(f)
                    logger.info(f'loader: {identifier} loaded into memory')
                except tomllib.TOMLDecodeError:
                    logger.error(f'Invalid TOML config: configs/{filename.stem}')

    target.configs = configs
    return target.configs



import logging
import tkinter as tk
from functools import cached_property
from pathlib import Path
from tkinter import messagebox, ttk
from typing import NamedTuple

import voicemeeterlib
from voicemeeterlib import kinds

from .builders import MainFrameBuilder
from .configurations import loader
from .data import _base_values, _configuration, get_configuration
from .errors import VMCompactError
from .menu import Menus
from .subject import Subject

logger = logging.getLogger(__name__)


class App(tk.Tk):
    """App mainframe"""

    @classmethod
    def make(cls, kind: NamedTuple):
        """
        Factory function for App.

        Returns an App class of a kind.
        """

        APP_cls = type(
            f'Voicemeeter{kind}.Compact',
            (cls,),
            {
                'kind': kind,
            },
        )
        return APP_cls

    def __init__(self, vmr, theme):
        super().__init__()
        self.logger = logger.getChild(self.__class__.__name__)
        self._vmr = vmr
        self._vmr.event.add(['pdirty', 'ldirty'])
        self.subject = Subject()
        self.start_updates()
        self._vmr.init_thread()
        for pn in (
            Path(__file__).parent.resolve() / 'img' / 'cat.ico',
            Path.cwd() / '_internal' / 'img' / 'cat.ico',
        ):
            if pn.is_file():
                self.iconbitmap(str(pn))
                break
        self.minsize(275, False)
        self._configs = None
        self.protocol('WM_DELETE_WINDOW', self.on_close_window)
        self.menu = self['menu'] = Menus(self, vmr)
        self.styletable = ttk.Style()
        if _configuration.config:
            vmr.apply_config(_configuration.config)

        self.build_app()

        self.drag_id = ''
        self.bind('<Configure>', self.dragging)

        self.after(1, self.healthcheck_step)

    def __str__(self):
        return f'{type(self).__name__}App'

    @property
    def target(self):
        """returns the current interface"""

        return self._vban if _base_values.vban_connected else self._vmr

    @property
    def configframes(self):
        """returns the current configframes"""

        return (
            frame
            for frame in self.winfo_children()
            if isinstance(frame, ttk.Frame)
            and '!stripconfig' in str(frame)
            or '!busconfig' in str(frame)
        )

    def build_app(self, kind=None, vban=None):
        """builds the app frames according to a kind"""
        self._vban = vban
        if kind:
            self.kind = kind

        # register event callbacks
        self.target.subject.add([self.on_pdirty, self.on_ldirty])

        self.bus_frame = None
        self.submix_frame = None
        self.builder = MainFrameBuilder(self)
        self.builder.setup()
        self.builder.create_channelframe('strip')
        self.builder.create_separator()
        self.builder.create_navframe()
        if _configuration.extended:
            self.nav_frame.extend.set(True)
            self.nav_frame.extend_frame()
        if self.kind.name == 'potato':
            self.builder.create_banner()

    def on_pdirty(self):
        if _base_values.run_update:
            self.after(1, self.subject.notify, 'pdirty')

    def on_ldirty(self):
        if not _base_values.dragging:
            self.after(1, self.subject.notify, 'ldirty')

    def _destroy_top_level_frames(self):
        """
        Clear observables.

        Deregister app as observer.

        Destroy all top level frames.
        """
        self.target.subject.remove([self.on_pdirty, self.on_ldirty])
        self.subject.clear()
        [
            frame.destroy()
            for frame in self.winfo_children()
            if isinstance(frame, ttk.Frame)
        ]

    def dragging(self, event, *args):
        if event.widget is self:
            if self.drag_id == '':
                _base_values.dragging = True
            else:
                self.after_cancel(self.drag_id)
            self.drag_id = self.after(100, self.stop_drag)

    def stop_drag(self):
        self.drag_id = ''
        _base_values.dragging = False

    @cached_property
    def userconfigs(self):
        self._configs = loader(self.kind.name, self.target)
        return self._configs

    def start_updates(self):
        def init():
            self.logger.debug('updates started')
            _base_values.run_update = True

        if self._vmr.gui.launched_by_api:
            self.subject.notify('pdirty')
            self.after(12000, init)
        else:
            init()

    def healthcheck_step(self):
        if not _base_values.vban_connected:
            try:
                self._vmr.version
            except voicemeeterlib.error.CAPIError:
                resp = messagebox.askyesno(message='Restart Voicemeeter GUI?')
                if resp:
                    self.logger.debug(
                        'healthcheck failed, rebuilding the app after GUI restart.'
                    )
                    self._vmr.end_thread()
                    self._vmr.run_voicemeeter(self._vmr.kind.name)
                    _base_values.run_update = False
                    self._vmr.init_thread()
                    self.after(8000, self.start_updates)
                    self._destroy_top_level_frames()
                    self.build_app(self._vmr.kind)
                    vban_config = get_configuration('vban')
                    for i, _ in enumerate(vban_config):
                        target = getattr(self.menu, f'menu_vban_{i + 1}')
                        target.entryconfig(0, state='normal')
                        target.entryconfig(1, state='disabled')
                    [
                        self.menu.menu_vban.entryconfig(j, state='normal')
                        for j, _ in enumerate(self.menu.menu_vban.winfo_children())
                    ]
                else:
                    self.destroy()
        self.after(250, self.healthcheck_step)

    def on_close_window(self):
        if _base_values.vban_connected:
            self._vban.logout()
        self.destroy()


_apps = {kind.name: App.make(kind) for kind in kinds.all}


def connect(kind_id: str, vmr, theme=None) -> App:
    """return App of the kind requested"""

    try:
        VMMIN_cls = _apps[kind_id]
    except KeyError:
        raise VMCompactError(f'Invalid kind: {kind_id}')
    return VMMIN_cls(vmr, theme)



def get_busmode_fullnames(kind) -> dict:
    if kind.name == 'basic':
        return {
            'normal': 'Normal',
            'amix': 'Mix Down A',
            'repeat': 'Stereo Repeat',
            'composite': 'Composite',
        }
    return {
        'normal': 'Normal',
        'amix': 'Mix Down A',
        'bmix': 'Mix Down B',
        'repeat': 'Stereo Repeat',
        'composite': 'Composite',
        'tvmix': 'Up Mix TV',
        'upmix21': 'Up Mix 2.1',
        'upmix41': 'Up Mix 4.1',
        'upmix61': 'Up Mix 6.1',
        'centeronly': 'Center Only',
        'lfeonly': 'LFE Only',
        'rearonly': 'Rear Only',
    }


def get_busmode_fullnames_reversed(kind) -> dict:
    return {v: k for k, v in get_busmode_fullnames(kind).items()}


def get_busmode_shortnames(kind) -> list:
    return list(get_busmode_fullnames(kind).keys())


def get_busmono_modes() -> list:
    return ['Mono: off', 'Mono: on', 'Stereo Reverse']



import logging
from tkinter import ttk

from . import builders
from .data import _base_values, _configuration

logger = logging.getLogger(__name__)


class Config(ttk.Frame):
    def __init__(self, parent, index, _id):
        super().__init__(parent)
        self.parent = parent
        self.index = index
        self.id = _id
        self.logger = logger.getChild(self.__class__.__name__)
        self.styletable = parent.styletable
        self.phys_in, self.virt_in = parent.kind.ins
        self.phys_out, self.virt_out = parent.kind.outs

        self.parent.subject.add(self)

    @property
    def identifier(self):
        return self.id

    @property
    def target(self):
        """returns the current interface"""

        return self.parent.target

    def getter(self, param):
        param = param.split('.')
        try:
            if len(param) == 2:
                target = getattr(self.target, param[0])
                return getattr(target, param[1])
            else:
                return getattr(self.target, param[0])
        except AttributeError as e:
            self.logger.error(f'{type(e).__name__}: {e}')

    def setter(self, param, value):
        param = param.split('.')
        try:
            if len(param) == 2:
                target = getattr(self.target, param[0])
                setattr(target, param[1], value)
            else:
                setattr(self.target, param[0], value)
        except AttributeError as e:
            self.logger(f'{type(e).__name__}: {e}')

    def scale_press(self, *args):
        self.after(1, self.remove_events)

    def remove_events(self):
        self.parent.target.event.remove('pdirty')
        self.parent.target.event.remove('ldirty')

    def scale_release(self, *args):
        _base_values.run_update = False
        self.after(1, self.add_events)

    def add_events(self):
        self.parent.target.event.add('pdirty')
        self.parent.target.event.add('ldirty')
        self.after(350, self.resume_updates)

    def pause_updates(self, func, *args):
        """function wrapper, adds a 50ms delay on updates"""
        _base_values.run_update = False

        func(*args)

        self.after(50, self.resume_updates)

    def resume_updates(self):
        _base_values.run_update = True

    def scale_enter(self, param, *args):
        val = self.slider_vars[self.slider_params.index(param)].get()
        self.parent.nav_frame.info_text.set(round(val, 1))

    def scale_leave(self, *args):
        self.parent.nav_frame.info_text.set('')

    def scale_callback(self, param, *args):
        """callback function for scale widget"""

        val = self.slider_vars[self.slider_params.index(param)].get()
        self.setter(param, round(val, 1))
        self.parent.nav_frame.info_text.set(round(val, 1))

    def reset_scale(self, param, val, *args):
        self.setter(param, val)
        self.slider_vars[self.slider_params.index(param)].set(val)

    def toggle_p(self, param):
        val = self.bool_param_vars[self.bool_params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f'{param}.TButton', background=f'{"green" if val else "white"}'
            )

    def on_update(self, subject):
        """update parameters"""
        if subject == 'pdirty':
            self.sync()


class StripConfig(Config):
    def __init__(self, parent, index, _id):
        super().__init__(parent, index, _id)
        self.grid(column=0, row=1, columnspan=4, padx=(2,))
        self.builder = builders.StripConfigFrameBuilder(self)
        self.builder.setup()
        self.make_row_0()
        self.make_row_1()
        self.make_row_2()
        self.builder.grid_configure()

        self.parent.target.clear_dirty()
        self.sync()

    @property
    def target(self):
        """returns the strip class for this configframe in the current interface"""

        _target = super(StripConfig, self).target
        return getattr(_target, self.identifier)[self.index]

    def make_row_0(self):
        if self.index < self.phys_in:
            if self.parent.kind.name == 'basic':
                self.builder.create_audibility_slider()
            else:
                self.builder.create_comp_slider()
                self.builder.create_gate_slider()
                self.builder.create_limit_slider()

    def make_row_1(self):
        self.builder.create_a_buttons()
        self.builder.create_b_buttons()

    def make_row_2(self):
        self.builder.create_param_buttons()

    def toggle_a(self, param):
        val = self.phys_out_params_vars[self.phys_out_params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f'{param}.TButton', background=f'{"green" if val else "white"}'
            )

    def toggle_b(self, param):
        val = self.virt_out_params_vars[self.virt_out_params.index(param)].get()
        self.setter(param, val)
        if not _configuration.themes_enabled:
            self.styletable.configure(
                f'{param}.TButton', background=f'{"green" if val else "white"}'
            )

    def teardown(self):
        self.builder.teardown()

    def sync(self):
        [
            self.phys_out_params_vars[i].set(self.getter(param))
            for i, param in enumerate(self.phys_out_params)
        ]
        [
            self.virt_out_params_vars[i].set(self.getter(param))
            for i, param in enumerate(self.virt_out_params)
        ]
        [
            self.bool_param_vars[i].set(self.getter(param))
            for i, param in enumerate(self.bool_params)
        ]
        [
            self.slider_vars[i].set(self.getter(param))
            for i, param in enumerate(self.slider_params)
            if self.index < self.phys_in
        ]

        if not _configuration.themes_enabled:
            [
                self.styletable.configure(
                    f'{param}.TButton',
                    background=f'{"green" if self.phys_out_params_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.phys_out_params)
            ]
            [
                self.styletable.configure(
                    f'{param}.TButton',
                    background=f'{"green" if self.virt_out_params_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.virt_out_params)
            ]
            [
                self.styletable.configure(
                    f'{param}.TButton',
                    background=f'{"green" if self.param_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.bool_params)
            ]


class BusConfig(Config):
    def __init__(self, parent, index, _id):
        super().__init__(parent, index, _id)
        if _configuration.extends_horizontal:
            self.grid(column=0, row=1, columnspan=4, padx=(2,))
        else:
            self.grid(column=0, row=3, columnspan=4, padx=(2,))
        self.builder = builders.BusConfigFrameBuilder(self, parent)
        self.builder.setup()
        self.make_row_0()
        self.make_row_1()
        self.builder.grid_configure()

        self.parent.target.clear_dirty()
        self.sync()

    @property
    def target(self):
        """returns the bus class for this configframe in the current interface"""

        _target = super(BusConfig, self).target
        return getattr(_target, self.identifier)[self.index]

    def make_row_0(self):
        self.builder.create_bus_mode_button()

    def make_row_1(self):
        self.builder.create_bus_mono_button()
        self.builder.create_param_buttons()

    def current_bus_mode(self):
        return self.target.mode.get()

    def rotate_bus_modes_right(self, *args):
        current_mode = self.bus_mode_map_reverse[self.bus_mode_label_text.get()]
        current_index = self.bus_modes.index(current_mode)
        next_index = (current_index + 1) % len(self.bus_modes)
        next_mode = self.bus_modes[next_index]

        setattr(self.target.mode, next_mode, True)
        self.bus_mode_label_text.set(self.bus_mode_map[next_mode])

    def rotate_bus_modes_left(self, *args):
        current_mode = self.bus_mode_map_reverse[self.bus_mode_label_text.get()]
        current_index = self.bus_modes.index(current_mode)
        prev_index = (current_index - 1) % len(self.bus_modes)
        prev_mode = self.bus_modes[prev_index]

        setattr(self.target.mode, prev_mode, True)
        self.bus_mode_label_text.set(self.bus_mode_map[prev_mode])

    def rotate_mono_right(self, *args):
        current_val = self.mono_modes.index(self.bus_mono_label_text.get())
        next_val = (current_val + 1) % 3
        self.bus_mono_label_text.set(self.mono_modes[next_val])
        self.setter('mono', next_val)

    def rotate_mono_left(self, *args):
        current_val = self.mono_modes.index(self.bus_mono_label_text.get())
        next_val = (current_val - 1) % 3
        self.bus_mono_label_text.set(self.mono_modes[next_val])
        self.setter('mono', next_val)

    def teardown(self):
        self.builder.teardown()

    def sync(self):
        [
            self.bool_param_vars[i].set(self.getter(param))
            for i, param in enumerate(self.bool_params)
        ]
        self.bus_mode_label_text.set(self.bus_mode_map[self.current_bus_mode()])
        if not _configuration.themes_enabled:
            [
                self.styletable.configure(
                    f'{param}.TButton',
                    background=f'{"green" if self.bool_param_vars[i].get() else "white"}',
                )
                for i, param in enumerate(self.bool_params)
            ]



from dataclasses import dataclass

from voicemeeterlib import kinds

from .configurations import get_configuration

configuration = get_configuration('app')


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]


@dataclass
class Configurations(metaclass=SingletonMeta):
    # is the gui extended
    extended: bool = configuration['extends']['extended']
    # direction the gui extends
    extends_horizontal: bool = configuration['extends']['extends_horizontal']
    # are themes enabled
    themes_enabled: bool = configuration['theme']['enabled']
    # light or dark
    theme_mode: str = configuration['theme']['mode']
    # size of mousewheel scroll step
    mwscroll_step: int = configuration['mwscroll_step']['size']
    # bus assigned as current submix
    submixes: int = configuration['submixes']['default']

    # width of a single channel labelframe
    channel_width: int = configuration['channel']['width']
    # height of a single channel labelframe
    channel_height: int = configuration['channel']['height']
    # xpadding for a single channel labelframe
    channel_xpadding: int = configuration['channel']['xpadding']

    # do we grid the navigation frame?
    navigation_show: bool = configuration['navigation']['show']

    @property
    def config(self):
        if 'configs' in configuration:
            return configuration['configs']['config']


@dataclass
class BaseValues(metaclass=SingletonMeta):
    # pause updates after releasing scale
    run_update: bool = False
    # are we dragging main window with mouse 1
    dragging: bool = False
    # a vban connection established
    vban_connected: bool = False


_base_values = BaseValues()
_configuration = Configurations()


def kind_get(kind_id):
    return kinds.request_kind_map(kind_id)
